// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"fmt"
	"os"

	"github.com/m4rw3r/uuid"
	"github.com/spf13/cobra"

	"bitbucket.org/orus-io/go-xbus/api"
)

// actorCmd represents the actor command
var actorCmd = &cobra.Command{
	Use:   "actor",
	Short: "A brief description of your command",
	Long: `A longer description that spans multiple lines and likely contains examples
and usage of using your command. For example:

Cobra is a CLI library for Go that empowers applications.
This application is a tool to generate the needed files
to quickly create a Cobra application.`,
	//Run: func(cmd *cobra.Command, args []string) {
	// TODO: Work your own magic here
	//fmt.Println("actor called")
	//},
}

// listCmd represents the list command
var actorCreateCmd = &cobra.Command{
	Use:   "create",
	Short: "Create a new actor",
	Long:  ``,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient()
		if err != nil {
			return
		}
		id, _ := uuid.FromString(cmd.Flag("id").Value.String())
		actor := api.Actor{
			ID:    id,
			Name:  cmd.Flag("name").Value.String(),
			Roles: []string{cmd.Flag("type").Value.String()},
		}
		err = actor.Kind.UnmarshalText([]byte(cmd.Flag("type").Value.String()))
		if err != nil {
			Logger.Errorf("Error parsing type. Got %s", err)
			return
		}
		created, err := c.ActorCreate([]api.Actor{actor})
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		if len(created) == 0 {
			fmt.Println("No actor created")
		} else {
			for _, actor := range created {
				fmt.Println(fmtActor("", actor))
			}
		}
	},
}

// listCmd represents the list command
var actorListCmd = &cobra.Command{
	Use:   "list",
	Short: "",
	Long:  ``,
	Run: func(cmd *cobra.Command, args []string) {
		c, err := GetClient()
		if err != nil {
			return
		}
		id, _ := uuid.FromString(cmd.Flag("id").Value.String())
		actor := api.Actor{
			ID:    id,
			Name:  cmd.Flag("name").Value.String(),
			Roles: []string{cmd.Flag("type").Value.String()},
		}
		if cmd.Flag("type").Value.String() != "" {
			err = actor.Kind.UnmarshalText([]byte(cmd.Flag("type").Value.String()))
			if err != nil {
				Logger.Errorf("Error parsing type. Got %s", err)
				return
			}
		}
		actors, err := c.ActorList([]api.Actor{actor})
		if err != nil {
			Logger.Errorf("Error: %s", err)
			return
		}
		if len(actors) == 0 {
			fmt.Println("No actors found")
		} else {
			for _, actor := range actors {
				fmt.Println(fmtActor("", actor))
			}
		}
	},
}

// actorAcceptCmd represents the actor accept sub-command.
var actorAcceptCmd = &cobra.Command{
	Use:   "accept",
	Short: "Accept an actor.",
	Long: "Accept a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: acceptActor,
}

// acceptActor parses arguments to accept a pending actor.
func acceptActor(cmd *cobra.Command, args []string) {
	actorID := cmd.Flag("id").Value.String()
	os.Exit(doAcceptActor(actorID))
}

// doAcceptActor accepts a pending actor request.
func doAcceptActor(actorID string) int {
	if len(actorID) == 0 {
		Logger.Errorf("Please provide an actor ID.")
		return 1
	}

	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	// TODO Find the actor based on the name.

	// Send an "accept" request.
	repActor, err := conn.ActorAccept(uuid.MustFromString(actorID))
	if err != nil {
		Logger.Errorf("Could not accept the actor %q: %s", actorID, err)
		return 1
	}
	if repActor.Status == api.ActorPending {
		Logger.Errorf("Could not accept the actor %q: Still pending.", actorID)
		return 1
	}
	Logger.Noticef("Successfully accepted the actor %q (ID %q).", repActor.Name, actorID)
	return 0
}

// actorRejectCmd represents the actor reject sub-command.
var actorRejectCmd = &cobra.Command{
	Use:   "reject",
	Short: "Reject an actor.",
	Long: "Reject a pending actor so it can be allowed to connect to Xbus. An actor ID or name " +
		"is required.",
	Run: rejectActor,
}

// rejectActor parses arguments to reject a pending actor.
func rejectActor(cmd *cobra.Command, args []string) {
	actorID := cmd.Flag("id").Value.String()
	os.Exit(doRejectActor(actorID))
}

// doRejectActor rejects a pending actor request.
func doRejectActor(actorID string) int {
	if len(actorID) == 0 {
		Logger.Errorf("Please provide an actor ID.")
		return 1
	}

	conn, err := GetClient()
	if err != nil {
		Logger.Errorf("Cannot connect to the NATS server: %s", err)
		return 1
	}

	// TODO Find the actor based on the name.

	// Send an "reject" request.
	repActor, err := conn.ActorReject(uuid.MustFromString(actorID))
	if err != nil {
		Logger.Errorf("Could not reject the actor %q: %s", actorID, err)
		return 1
	}
	if repActor.Status == api.ActorPending {
		Logger.Errorf("Could not reject the actor %q: Still pending.", actorID)
		return 1
	}
	Logger.Noticef("Successfully rejected the actor %q (ID %q).", repActor.Name, actorID)
	return 0
}

func fmtActor(prefix string, actor api.Actor) string {
	s := fmt.Sprintf(
		"%sID: %s\n%sName: %s\n%sKind: %s\n%sRole: %s\n",
		prefix, actor.ID,
		prefix, actor.Name,
		prefix, actor.Kind,
		prefix, actor.Roles[0])
	return s
}

func init() {
	RootCmd.AddCommand(actorCmd)
	actorCmd.AddCommand(actorCreateCmd)
	actorCmd.AddCommand(actorListCmd)
	actorCmd.AddCommand(actorAcceptCmd)
	actorCmd.AddCommand(actorRejectCmd)

	flags := actorCmd.PersistentFlags()
	flags.String("id", "", "Actor's ID")
	flags.String("name", "", "Actor's name")
	flags.String("role", "", "Actor's role")
	flags.String("type", "", "Actor's type (emitter, consumer)")
}
