// Copyright 2016 Orus.io. All rights reserved.

package cmd

import (
	"crypto/x509"
	"encoding/pem"
	"io/ioutil"
	"os"

	"github.com/spf13/cobra"
	"github.com/spf13/viper"

	"bitbucket.org/orus-io/go-xbus/tools"
)

// certCmd represents the certificate command.
var certCmd = &cobra.Command{
	Use:   "cert",
	Short: "Manage certificates used by this program.",
	Long:  "Manage certificates used by this program.",
}

// certGenerateCmd represents the certificate generation sub-command.
var certGenerateCmd = &cobra.Command{
	Use:   "generate",
	Short: "Generate certificates.",
	Long: "Generate certificates to be used by this program, based on private keys; see " +
		"sub-commands for details on the certificates that can be generated here. Also check " +
		"the \"key\" command for private key generation.",
}

// certGenerateServerCACmd represents the server CA generation sub-command.
var certGenerateServerCACmd = &cobra.Command{
	Use:   "server-ca",
	Short: "Generate certificates for the server CA.",
	Long: "Generate certificates for the server CA (the one clients connecting to Xbus will " +
		"check). The file will be named \"ServerCA.crt\" and is based on a \"ServerCA.key\" " +
		"private key.",
	Run: generateServerCA,
}

// generateServerCA parses arguments to generate the server CA.
func generateServerCA(cmd *cobra.Command, args []string) {
	certDir := viper.GetString("cert-dir")
	hostDefs := cmd.Flag("host-defs").Value.String()
	if hostDefs == "" {
		hostDefs = viper.GetString("nats.host")
	}
	os.Exit(doGenerateServerCA(certDir, hostDefs))
}

// doGenerateServerCA generates the server CA.
func doGenerateServerCA(certDir string, hostDefs string) int {
	generated := genCert(
		certDir+"/ServerCA.crt", certDir+"/ServerCA.key",
		func(priv interface{}) ([]byte, error) {
			return tools.CreateServerRootCA(priv, hostDefs)
		},
	)
	if !generated {
		return 1
	}
	return 0
}

// certGenerateClientCACmd represents the client CA generation sub-command.
var certGenerateClientCACmd = &cobra.Command{
	Use:   "client-ca",
	Short: "Generate certificates for the client CA.",
	Long: "Generate certificates for the client CA (the one clients connecting to Xbus will " +
		"have to have been signed by). The file will be named \"ClientCA.crt\" and is based on " +
		"a \"ClientCA.key\" private key.",
	Run: generateClientCA,
}

// generateClientCA parses arguments to generate the client CA.
func generateClientCA(cmd *cobra.Command, args []string) {
	certDir := viper.GetString("cert-dir")
	os.Exit(doGenerateClientCA(certDir))
}

// doGenerateClientCA generates the client CA.
func doGenerateClientCA(certDir string) int {
	generated := genCert(
		certDir+"/ClientCA.crt", certDir+"/ClientCA.key",
		tools.CreateClientCA,
	)
	if !generated {
		return 1
	}
	return 0
}

func init() {
	flags := certGenerateServerCACmd.PersistentFlags()
	flags.String(
		"host-defs", "",
		"A host defs for the server certificate. If ommited, --nats-host is used")

	RootCmd.AddCommand(certCmd)

	certCmd.AddCommand(certGenerateCmd)

	certGenerateCmd.AddCommand(certGenerateServerCACmd)

	certGenerateCmd.AddCommand(certGenerateClientCACmd)
}

// loadPrivKey loads the private key stored in the specified file.
func loadPrivKey(privKeyPath string) (privKey interface{}, err error) {
	privKeyData, err := ioutil.ReadFile(privKeyPath)
	if err != nil {
		return nil, err
	}
	privKeyBlock, _ := pem.Decode(privKeyData)
	return x509.ParseECPrivateKey(privKeyBlock.Bytes)
}

// genCert is a helper that creates a certificate and saves it into a file.
func genCert(
	certPath string, privKeyPath string,
	genFunc func(priv interface{}) (certData []byte, err error),
) (generated bool) {

	// Load the private key.
	privKey, err := loadPrivKey(privKeyPath)
	if err != nil {
		Logger.Errorf("Cannot load the %q private key", privKeyPath)
		return false
	}

	// Generate the certificate.
	certData, err := genFunc(privKey)
	if err != nil {
		Logger.Errorf("Cannot create the certificate: %s", err)
		return false
	}
	if err := ioutil.WriteFile(certPath, certData, 0644); err != nil {
		Logger.Errorf("Cannot save the certificate to %q: %s", certPath, err)
		return false
	}

	Logger.Noticef("Successfully generated %q.", certPath)
	return true
}
