// Copyright 2016 Orus.io. All rights reserved.

package fullstack_test

import (
	"testing"
	"time"

	_ "github.com/mattn/go-sqlite3"

	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/testutils"
)

func TestStartupShutdown(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	defer ft.Shutdown()
}

func TestEndToEndSingleEvent(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")

	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit <- e
			consumer.ProcessingSuccess(e)
		})
	assert.Nil(t, err)
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(BuildSimpleGraph("evt1", "anyconsumer"))
	assert.Nil(t, err)

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan := make(chan api.EmitterEnvelopeState)
	sub, err := emitter.GetClient().GetConn().BindRecvChan(
		api.EmitterEnvelopeStateSubject(emitter.ID, env.ID), envStateChan)
	if err != nil {
		t.Fatal(err)
	}
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	timeout := testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case envState := <-envStateChan:
		assert.Equal(t, api.ProcessRunning, envState.Status)
	}

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case envState := <-envStateChan:
		assert.Equal(t, api.ProcessDone, envState.Status)
	}
}

func TestEndToEndSingleEventWithConsumerDown(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	var handler = func(input string, r *xbus.EnvelopeReader) {
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return
		}
		hit <- e
		consumer.ProcessingSuccess(e)
	}
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(BuildSimpleGraph("evt1", "anyconsumer"))
	assert.Nil(t, err)

	// take the consumer offline
	xsub.Unsubscribe()

	ft.Logger.ExpectsError("[server] [emitter] Actor did not ack the envelops. Mark it unresponsivs.")

	status, err := emitter.SendItems(api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	// wait enough for the consumer to be declared 'off', which is a 50 milliseconds timeout (in tests)
	timeout := testutils.TimeoutChan(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Visible)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	pList, err := ft.Client.ProcessQuery(false)
	assert.Nil(t, err)
	assert.Equal(t, 1, len(pList))
	assert.Equal(t, api.ProcessRunning, pList[0].Status)

	// go back online
	xsub, err = consumer.Subscribe(handler)
	assert.Nil(t, err)

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithServerShutdown(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")
	xbusClient, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	handler := func(input string, r *xbus.EnvelopeReader) {
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return
		}
		hit <- e
		consumer.ProcessingSuccess(e)
	}
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(BuildSimpleGraph("evt1", "anyconsumer"))
	assert.Nil(t, err)

	// take the consumer offline
	xsub.Unsubscribe()
	xbusClient.Close()

	ft.Logger.ExpectsError("[server] [emitter] Actor did not ack the envelops. Mark it unresponsivs.")

	status, err := emitter.SendItems(api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	// wait enough for the consumer to be declared 'off', which is a 50 milliseconds timeout (in tests)
	timeout := testutils.TimeoutChan(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Visible)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	// stop/start the server
	ft.ServerStop()
	assert.True(t, ft.ServerStart())

	// go back online
	assert.Nil(t, xbusClient.Connect())
	xsub, err = consumer.Subscribe(handler)
	assert.Nil(t, err)

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithPauseResume(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	handler := func(input string, r *xbus.EnvelopeReader) {
		e, err := r.Complete(time.Second)
		if err != nil {
			ft.T.Error(err)
			return
		}
		hit <- e
		consumer.ProcessingSuccess(e)
	}
	xsub, err := consumer.Subscribe(handler)
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(BuildSimpleGraph("evt1", "anyconsumer"))
	assert.Nil(t, err)

	// take the consumer offline
	xsub.Unsubscribe()

	ft.Logger.ExpectsError("[server] [emitter] Actor did not ack the envelops. Mark it unresponsivs.")

	status, err := emitter.SendItems(api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	// wait enough for the consumer to be declared 'off', which is a 50 milliseconds timeout (in tests)
	timeout := testutils.TimeoutChan(300 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Visible)
	case <-hit:
		t.Fatalf("Very wrong, we unsubscribed the consumer, we should not receive anything (yet)")
	}

	pList, err := ft.Client.ProcessQuery(false)
	assert.Nil(t, err)
	assert.Equal(t, 1, len(pList))
	assert.Equal(t, api.ProcessRunning, pList[0].Status)

	err = ft.Client.ProcessControl(pList[0].ID, "pause")
	assert.Nil(t, err)

	// go back online
	xsub, err = consumer.Subscribe(handler)
	assert.Nil(t, err)

	timeout = testutils.TimeoutChan(500 * time.Millisecond)
	select {
	case <-timeout:
		// everything is normal
		actor, exists := ft.Server.ActorStorage().Get(consumer.ID)
		assert.True(t, exists)
		assert.False(t, actor.Visible)
	case <-hit:
		t.Fatalf("Very wrong, we paused the process, we should not receive anything")
	}

	err = ft.Client.ProcessControl(pList[0].ID, "resume")
	assert.Nil(t, err)

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}
}

func TestEndToEndSingleEventWithConsumerError(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	xsub, err := consumer.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit <- e
			consumer.ProcessingError(e, "bad thing happened: %s", "I just want not to work")
		})
	if err != nil {
		t.Fatal(err)
	}
	defer xsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(BuildSimpleGraph("evt1", "anyconsumer"))
	assert.Nil(t, err)

	env := xbus.ForgeEnvelopeFromItems("evt1", []api.Item{[]byte("\"item1\"")})

	envStateChan := make(chan api.EmitterEnvelopeState)
	sub, err := emitter.GetClient().GetConn().BindRecvChan(
		api.EmitterEnvelopeStateSubject(emitter.ID, env.ID), envStateChan)
	if err != nil {
		t.Fatal(err)
	}
	defer sub.Unsubscribe()

	status, err := emitter.SendEnvelope(api.DefaultOutput, env)
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	timeout := testutils.TimeoutChan(5 * time.Second)
	var processID uuid.UUID
	select {
	case <-timeout:
		t.Errorf("Process not running")
	case envState := <-envStateChan:
		assert.Equal(t, api.ProcessRunning, envState.Status)

		for processID = range envState.ProcessStatusList {
		}
	}

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("Process not ended")
	case envState := <-envStateChan:
		assert.Equal(t, api.ProcessError, envState.Status)
		assert.Equal(t,
			"bad thing happened: I just want not to work",
			envState.Errors[processID][0].Text)

	}
}
