package fullstack_test

import (
	"encoding/json"
	"fmt"
	"testing"
	"time"

	_ "github.com/mattn/go-sqlite3"

	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

func forwardWorker(ft *FullstackEnv, name string) (*xbus.Client, *xbus.Actor, *xbus.Subscription) {
	client, worker := ft.CreateWorker(name, []string{"forward"})
	sub, err := worker.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			env, err := r.Complete(time.Second)
			if err != nil {
				ft.Logger.Errorf("Error in forwardWorker: %s", err)
			}
			newEnv := api.Envelope{
				ID:      env.ID,
				Context: env.Context,
				Last:    true,
			}
			newEnv.Context = env.Context
			worker.Log.EnvNoticef(env, "Forwarding the envelope")
			_, err = worker.SendEnvelope(
				api.DefaultOutput,
				newEnv)
			if err != nil {
				ft.Logger.Errorf("Error in forwardWorker: %s", err)
			}
		})
	if err != nil {
		panic(err)
	}
	return client, worker, sub
}

func TestOneWorker(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit := make(chan *api.Envelope, 1)

	ft.CreateRole(api.ConsumerKind, "anyconsumer")
	_, consumer := ft.CreateConsumer("jojo", []string{"anyconsumer"})
	csub, err := consumer.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit <- e
			consumer.ProcessingSuccess(e)
		})
	if err != nil {
		t.Fatal(err)
	}
	defer csub.Unsubscribe()

	_, worker, wsub := forwardWorker(ft, "fw")
	defer wsub.Unsubscribe()

	assert.Equal(t, "jojo", consumer.Name)

	_, _, err = ft.Client.GraphPush(fmt.Sprintf(`
name: forward
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: relay
  type: worker
  actorids: [%s]
  inputs: [default]
  outputs: [default]
- id: target
  type: consumer
  roles: [anyconsumer]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> relay
- relay.default -> target.default
`, worker.ID))
	assert.Nil(t, err)

	sub, err := ft.Client.GetConn().Conn.SubscribeSync(api.CoreProcessEventsSubject)
	assert.Nil(t, err)

	status, err := emitter.SendItems(api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	// Wait for a process to change status to 'done' (ours)
	for true {
		msg, err := sub.NextMsg(time.Second)
		assert.Nil(t, err)
		if err != nil {
			t.Fatal("No process done")
		}

		var msgDec api.CoreEvent
		err = json.Unmarshal(msg.Data, &msgDec)
		assert.Nil(t, err)
		if msgDec.Process.Status == api.ProcessDone {
			break
		}
	}

	var lastEnvelope *api.Envelope

	timeout := testutils.TimeoutChan(5 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope = <-hit:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)

		pState := ft.Server.ProcessStorage().GetState(lastEnvelope.Context.ProcessID)

		for node, state := range pState.NodeStates {
			fmt.Println(" -", node, ": ")
			fmt.Println("  ", state)
		}

		assert.Equal(t, api.ProcessDone, pState.Status)
	}

	// Make sure the log message landed correctly
	var logEntry model.LogEntry
	q := ft.DB.Query(ft.Model.LogEntry)
	q = q.Where(ft.Model.LogEntry.EnvelopeID.Eq(lastEnvelope.ID))
	assert.Nil(t, q.One(&logEntry))
	assert.Equal(t, lastEnvelope.ID, uuid.UUID(logEntry.EnvelopeID))
}

func routerWorker(ft *FullstackEnv, name string) (*xbus.Client, *xbus.Actor, *xbus.Subscription) {
	c, worker := ft.CreateWorker(name, []string{"router"})
	sub, err := worker.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			env, err := r.Complete(time.Second)
			if err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
			}
			output := ""
			closeOutput := ""
			if string(env.Events[0].Items[0]) == "\"item1\"" {
				output = "output_1"
				closeOutput = "output_2"
			} else {
				output = "output_2"
				closeOutput = "output_1"
			}
			_, err = worker.SendEnvelope(output, api.ForwardEnvelope(env))
			if err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
			}
			_, err = worker.SendEnvelope(closeOutput, api.CloseOutput(env.Context))
			if err != nil {
				ft.Logger.Errorf("Error in routerWorker: %s", err)
			}
			// Will close automatically output_3
			worker.ProcessingSuccess(env)
		},
	)
	if err != nil {
		panic(err)
	}
	return c, worker, sub
}

func TestRouting(t *testing.T) {
	ft := InitFullstackEnv(t, true)
	if ft == nil {
		return
	}
	defer ft.Shutdown()

	ft.CreateRole(api.EmitterKind, "anyemitter")
	_, emitter := ft.CreateEmitter("bob", []string{"anyemitter"})

	assert.Equal(t, "bob", emitter.Name)

	hit1 := make(chan *api.Envelope, 1)
	hit2 := make(chan *api.Envelope, 1)

	_, consumer1 := ft.CreateConsumer("jojo", []string{"consumerkind1"})
	sub1, err := consumer1.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit1 <- e
			consumer1.ProcessingSuccess(e)
		})
	if err != nil {
		t.Fatal(err)
	}

	defer sub1.Unsubscribe()

	_, consumer2 := ft.CreateConsumer("bugs", []string{"consumerkind2"})
	sub2, err := consumer2.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit2 <- e
			consumer2.ProcessingSuccess(e)
		})
	if err != nil {
		t.Fatal(err)
	}
	defer sub2.Unsubscribe()

	_, consumer3 := ft.CreateConsumer("last", []string{"consumerkind3"})
	sub3, err := consumer3.Subscribe(
		func(input string, r *xbus.EnvelopeReader) {
			e, err := r.Complete(time.Second)
			if err != nil {
				ft.T.Error(err)
				return
			}
			hit2 <- e
			consumer3.ProcessingSuccess(e)
		})
	if err != nil {
		t.Fatal(err)
	}
	defer sub3.Unsubscribe()

	_, _, subw := routerWorker(ft, "router-1")
	defer subw.Unsubscribe()

	_, _, err = ft.Client.GraphPush(fmt.Sprintf(`
name: routing
version: 0.1
status: active
nodes:
- id: source
  type: emitter
  sourcematch:
    eventtypes: [evt1]
  outputs: [default]
- id: router
  type: worker
  actors: [router-1]
  inputs: [default]
  outputs: [output_1, output_2, output_3]
- id: target1
  type: consumer
  roles: [consumerkind1]
  rolebroadcast: true
  inputs: [default]
- id: target2
  type: consumer
  roles: [consumerkind2]
  rolebroadcast: true
  inputs: [default]
- id: target3
  type: consumer
  roles: [consumerkind3]
  rolebroadcast: true
  inputs: [default]
edges:
- source -> router
- router.output_1 -> target1.default
- router.output_2 -> target2.default
- router.output_3 -> target3.default
`))
	assert.Nil(t, err)

	var pStateChan = make(chan api.EmitterEnvelopeState)
	emitter.GetClient().GetConn().BindRecvChan(
		api.EmitterEnvelopeStateSubject(emitter.ID, uuid.UUID{}),
		pStateChan)

	status, err := emitter.SendItems(api.DefaultOutput, "evt1", []api.Item{[]byte("\"item1\"")})
	assert.Nil(t, err)
	assert.Equal(t, api.ReceptionAccepted, status)

	timeout := testutils.TimeoutChan(1 * time.Second)
	select {
	case <-timeout:
		t.Errorf("No envelope received")
	case lastEnvelope := <-hit1:
		assert.Equal(t, "evt1", lastEnvelope.Events[0].Type)
	}

	timeout = testutils.TimeoutChan(2 * time.Second)
	var pStatus api.ProcessStatus
	for pStatus != api.ProcessDone {
		select {
		case <-timeout:
			t.Fatalf("Process never ended")
		case <-hit2:
			t.Fatalf("No envelope should have reached target_2/3")
		case eState := <-pStateChan:
			t.Logf("New envelope status: %s", eState.Status)
			pStatus = eState.Status
		}
	}
}
