package graph

import (
	"fmt"

	"bitbucket.org/orus-io/xbus/storage"
)

// ValidationErrorCode is a validation error code
type ValidationErrorCode int

const (
	// ErrMissingEmitter is when a graph has no emitter
	ErrMissingEmitter ValidationErrorCode = iota
	// ErrMissingConsumer is when a graph has no consumer
	ErrMissingConsumer
	// ErrExtraEmitter is when a graph has more than one emitter
	ErrExtraEmitter
	// ErrDisconnectedInput is when an input is not connected by any edge
	ErrDisconnectedInput
	// ErrDisconnectedOutput is when an ouput is not connected by any edge
	ErrDisconnectedOutput
	// ErrEdgeInvalidFrom is when the origin of an edge does not exist
	ErrEdgeInvalidFrom
	// ErrEdgeInvalidTo is when the target of an edge does not exist
	ErrEdgeInvalidTo
	// ErrCycle is when the graph has a cycle
	ErrCycle
	// ErrDisconnectedGraph is when the graph is not connected
	ErrDisconnectedGraph
	// ErrSourceIsNotEmitter is when a source node is not an emitter
	ErrSourceIsNotEmitter
	// ErrSinkIsNotConsumer is when a sink node is not a consumer
	ErrSinkIsNotConsumer

	// ErrInvalidNodeID is when a node ID is empty or contains any non [a-zA-Z0-9_\-] character
	ErrInvalidNodeID
	// ErrDuplicateNodeID is when two nodes have the same ID
	ErrDuplicateNodeID

	// ErrUnknownActor is when a node actorID item does not exist
	ErrUnknownActor
	// ErrNoActorForRole is when a role matches no actor
	ErrNoActorForRole

	// ErrTooManyRoles is when a node has more than 1 role
	ErrTooManyRoles
	// ErrRoleBroadcastUnset is when a role is set but rolebroadcast is not (it should)
	ErrRoleBroadcastUnset
)

// ValidationError is a detailed graph validation error
type ValidationError struct {
	Code      ValidationErrorCode
	NodeID    string // The node having an error if relevant
	InputRef  storage.NodeInputRef
	OutputRef storage.NodeOutputRef
	Edge      storage.Edge
	Path      Path // The path having an error if relevant
	Actor     string
	Role      string
}

func (v ValidationError) Error() string {
	switch v.Code {
	case ErrMissingEmitter:
		return "Graph must have one emitter (found none)"
	case ErrMissingConsumer:
		return "Graph must have at least one consumer (found none)"
	case ErrExtraEmitter:
		return fmt.Sprintf("Found extra emitter node: %s", v.NodeID)
	case ErrDisconnectedInput:
		return fmt.Sprintf("Input %s is not connected by any edge", v.InputRef)
	case ErrDisconnectedOutput:
		return fmt.Sprintf("Output %s is not connected by any edge", v.OutputRef)
	case ErrEdgeInvalidFrom:
		return fmt.Sprintf("Edge %s origin does not exist", v.Edge)
	case ErrEdgeInvalidTo:
		return fmt.Sprintf("Edge %s destination does not exist", v.Edge)
	case ErrCycle:
		return fmt.Sprintf("Cycle detected: %s", v.Path)
	case ErrDisconnectedGraph:
		return "The graph is not connected"
	case ErrSourceIsNotEmitter:
		return fmt.Sprintf("Source node '%s' is not an emitter", v.NodeID)
	case ErrSinkIsNotConsumer:
		return fmt.Sprintf("Sink node '%s' is not a consumer", v.NodeID)
	case ErrInvalidNodeID:
		return fmt.Sprintf("Node ID '%s' is not a valid ID. It should match the regexp '%s'", v.NodeID, idRegexpS)
	case ErrDuplicateNodeID:
		return fmt.Sprintf("Two or more nodes have the same ID '%s'", v.NodeID)
	case ErrUnknownActor:
		return fmt.Sprintf("Node %s references an unknown actor: %s", v.NodeID, v.Actor)
	case ErrNoActorForRole:
		return fmt.Sprintf("Node %s uses role '%s' but no actor implements it", v.NodeID, v.Role)
	case ErrTooManyRoles:
		return fmt.Sprintf("Node '%s' has more than 1 role", v.NodeID)
	case ErrRoleBroadcastUnset:
		return fmt.Sprintf("Node '%s' should have rolebroadcast enabled", v.NodeID)
	default:
		panic(fmt.Sprintf("Invalid error code: %d", v.Code))
	}
}

// MissingEmitterError returns a ErrMissingEmitter error
func MissingEmitterError() ValidationError {
	return ValidationError{Code: ErrMissingEmitter}
}

// MissingConsumerError returns a ErrMissingConsumer error
func MissingConsumerError() ValidationError {
	return ValidationError{Code: ErrMissingConsumer}
}

// ExtraEmitterError returns a ErrExtraEmitter error
func ExtraEmitterError(nodeID string) ValidationError {
	return ValidationError{Code: ErrExtraEmitter, NodeID: nodeID}
}

// DisconnectedInputError returns a ErrDisconnectedInput error
func DisconnectedInputError(inputRef storage.NodeInputRef) ValidationError {
	return ValidationError{Code: ErrDisconnectedInput, InputRef: inputRef}
}

// DisconnectedOutputError returns a ErrDisconnectedOutput error
func DisconnectedOutputError(outputRef storage.NodeOutputRef) ValidationError {
	return ValidationError{Code: ErrDisconnectedOutput, OutputRef: outputRef}
}

// EdgeInvalidFromError returns a ErrEdgeInvalidFrom error
func EdgeInvalidFromError(edge storage.Edge) ValidationError {
	return ValidationError{Code: ErrEdgeInvalidFrom, Edge: edge}
}

// EdgeInvalidToError returns a ErrEdgeInvalidTo error
func EdgeInvalidToError(edge storage.Edge) ValidationError {
	return ValidationError{Code: ErrEdgeInvalidTo, Edge: edge}
}

// CycleError returns a ErrGraphCycle error
func CycleError(path Path) ValidationError {
	return ValidationError{Code: ErrCycle, Path: path}
}

// DisconnectedGraphError returns a ErrDisconnectedGraph error
func DisconnectedGraphError() ValidationError {
	return ValidationError{Code: ErrDisconnectedGraph}
}

// SourceIsNotEmitterError returns a ErrSourceIsNotEmitter error
func SourceIsNotEmitterError(nodeID string) ValidationError {
	return ValidationError{Code: ErrSourceIsNotEmitter, NodeID: nodeID}
}

// SinkIsNotConsumerError returns a ErrSinkIsNotConsumer error
func SinkIsNotConsumerError(nodeID string) ValidationError {
	return ValidationError{Code: ErrSinkIsNotConsumer, NodeID: nodeID}
}

// InvalidNodeID returns a ErrInvalidNodeID
func InvalidNodeID(nodeID string) ValidationError {
	return ValidationError{Code: ErrInvalidNodeID, NodeID: nodeID}
}

// DuplicateNodeIDError returns a ErrDuplicateNodeID
func DuplicateNodeIDError(nodeID string) ValidationError {
	return ValidationError{Code: ErrDuplicateNodeID, NodeID: nodeID}
}

// UnknownActorError returns a ErrUnknownActor error
func UnknownActorError(nodeID string, actor string) ValidationError {
	return ValidationError{Code: ErrUnknownActor, NodeID: nodeID, Actor: actor}
}

// NoActorForRoleError returns a ErrNoActorForRole error
func NoActorForRoleError(nodeID string, role string) ValidationError {
	return ValidationError{Code: ErrNoActorForRole, NodeID: nodeID, Role: role}
}

// TooManyRolesError returns a ErrTooManyRoles error
func TooManyRolesError(nodeID string) ValidationError {
	return ValidationError{Code: ErrTooManyRoles, NodeID: nodeID}
}

// RoleBroadcastUnsetError returns a ErrRoleBroadcastUnset error
func RoleBroadcastUnsetError(nodeID string) ValidationError {
	return ValidationError{Code: ErrRoleBroadcastUnset, NodeID: nodeID}
}

// ValidationErrorList is a list or validation errors
type ValidationErrorList []ValidationError

// Add appends an error to the list
func (l *ValidationErrorList) Add(err ValidationError) {
	*l = append(*l, err)
}
