package graph

import (
	"testing"

	"github.com/stretchr/testify/assert"
	"gopkg.in/yaml.v2"
)

func TestValidator(t *testing.T) {
	var ValidatorTests = []struct {
		name      string
		graph     []byte
		errorMsgs []string
	}{
		{"emptygraph", []byte(""), []string{
			"Graph must have one emitter (found none)",
			"Graph must have at least one consumer (found none)",
		}},

		{"disconnected output & no consumer", []byte(`
nodes:
  - id: src
    type: emitter
    outputs: [default]
`),
			[]string{
				"Graph must have at least one consumer (found none)",
				"Output src.default is not connected by any edge",
			},
		},

		{"disconnected input & no emitter", []byte(`
nodes:
  - id: out
    type: consumer
    inputs: [default]
`),
			[]string{
				"Graph must have one emitter (found none)",
				"Input out.default is not connected by any edge",
			},
		},

		{"invalid edges", []byte(`
nodes:
  - id: in
    type: emitter
    outputs: [default]
  - id: out
    type: consumer
    inputs: [default]
edges:
  - in.default -> out.default
  - in.output2 -> out.default
  - out.output1 -> in.default
`),
			[]string{
				"Edge in.output2->out.default origin does not exist",
				"Edge out.output1->in.default destination does not exist",
				"Edge out.output1->in.default origin does not exist",
			},
		},

		{"loop", []byte(`
nodes:
- id: in
  type: emitter
  outputs: [default]
- id: w
  type: worker
  inputs: [in]
  outputs: [out]
- id: out
  type: consumer
  inputs: [default]
edges:
- in.default -> w.in
- w.out -> w.in
- w.out -> out.default
`),
			[]string{
				"Cycle detected: [w.out->w.in]",
			},
		},

		{"disconnected graph", []byte(`
nodes:
- id: in1
  type: emitter
  outputs: [default]
- id: out1
  type: consumer
  inputs: [default]
- id: in2
  type: emitter
  outputs: [default]
- id: out2
  type: consumer
  inputs: [default]
edges:
- in1->out1
- in2->out2
`),
			[]string{
				"Found extra emitter node: in2",
				"The graph is not connected",
			},
		},
		{"roles/broadcast", []byte(`
nodes:
- id: in1
  type: emitter
  roles: [role1, role2]
  outputs: [default]
- id: out1
  type: consumer
  roles: [role1]
  rolebroadcast: false
  inputs: [default]
edges:
- in1->out1
`),
			[]string{
				"Node 'in1' has more than 1 role",
				"Node 'in1' should have rolebroadcast enabled",
				"Node 'out1' should have rolebroadcast enabled",
			},
		},
		{"invalid source/sink types", []byte(`
nodes:
- id: in1
  type: worker
  outputs: [default]
- id: out1
  type: worker
  inputs: [i1, i2]
- id: in2
  type: emitter
  outputs: [default]
- id: out2
  type: consumer
  inputs: [default]
edges:
- in1->out1.i1
- in2->out2
- in2->out1.i2
`),
			[]string{
				"Source node 'in1' is not an emitter",
				"Sink node 'out1' is not a consumer",
			},
		},
		{"invalid id/duplicate id", []byte(`
nodes:
- id: in 1
  type: emitter
  outputs: [default]
- id: worker#1
  type: worker
  inputs: [default]
  outputs: [default]
- id: w2
  type: worker
  inputs: [default]
  outputs: [default]
- id: w2
  type: worker
  inputs: [default]
  outputs: [default]
- id:
  type: consumer
  inputs: [default]
edges:
- in 1->worker#1
- worker#1->w2
- w2->.default
  `),
			[]string{
				"Node ID 'in 1' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Node ID 'worker#1' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Two or more nodes have the same ID 'w2'",
				"Node ID '' is not a valid ID. It should match the regexp '^[a-zA-Z0-9_\\-]+$'",
				"Cycle detected: [.->. .->. w2.default->.default]",
				"The graph is not connected",
			},
		},
	}
	for _, test := range ValidatorTests {
		t.Run(test.name, func(t *testing.T) {
			var g Graph
			if err := yaml.Unmarshal(test.graph, &g); err != nil {
				t.Fatal(err)
			}
			errors := g.Validate()
			assert.Equal(t, test.errorMsgs, errors.Messages())
		})
	}
}
