package process

import (
	"fmt"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/graph"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewManager creates a process.Manager
func NewManager(
	graphManager graph.Manager,
	actorStorage storage.ActorStorage,
	processStorage storage.ProcessStorage,
	logger xbus.Logger,
) *Manager {
	manager := Manager{
		GraphManager:   graphManager,
		ActorStorage:   actorStorage,
		ProcessStorage: processStorage,
		Log:            xbus.NewPrefixedLogger("[process/manager] ", logger),
	}
	processStorage.SetNextStateCallback(manager.nextState)
	return &manager
}

// Manager manage the processes
type Manager struct {
	GraphManager   graph.Manager
	ActorStorage   storage.ActorStorage
	ProcessStorage storage.ProcessStorage

	Log xbus.Logger

	CurrentProcessList []*Process
}

// Close unset callback on ProcessStorage
func (m *Manager) Close() {
	m.ProcessStorage.SetNextStateCallback(nil)
}

func (m *Manager) loadProcess(p storage.Process) (*Process, error) {
	g, exists := m.GraphManager.GetGraph(p.GraphID)
	if !exists {
		return nil, fmt.Errorf("No such graph: %s", p.GraphID)
	}
	return &Process{
		ID:                p.ID,
		actors:            m.ActorStorage,
		processStorage:    m.ProcessStorage,
		TriggerEmitterID:  p.TriggerEmitterID,
		TriggerEnvelopeID: p.TriggerEnvelopeID,
		Graph:             g,
		SourceRef:         p.SourceOutput,
	}, nil
}

// LookupOrCreate returns the Process handling the given envelope or create
// if necessary
func (m *Manager) LookupOrCreate(emitter api.Actor, envelopeID uuid.UUID, eventTypes []string) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if ok {
		return m.loadProcess(p)
	}
	graph, sourceOutput, ok := m.GraphManager.LookupGraph(emitter, eventTypes)
	if !ok {
		return nil, fmt.Errorf("No matching graph found")
	}

	p = storage.Process{
		TriggerEmitterID:  emitter.ID,
		TriggerEnvelopeID: envelopeID,
		GraphID:           graph.ID,
		SourceOutput:      sourceOutput,
	}
	var err error
	p.ID, err = m.ProcessStorage.CreateProcess(p)
	if err != nil {
		return nil, err
	}
	return m.loadProcess(p)
}

// GetEmitterEnvelopeStatus returns the statuses of the processes triggered
// by the given emitter/envelope
func (m *Manager) GetEmitterEnvelopeStatus(emitter api.Actor, envelopeID uuid.UUID) map[uuid.UUID]api.ProcessStatus {
	p, ok := m.ProcessStorage.GetProcessByTrigger(emitter, envelopeID)
	if !ok {
		return nil
	}
	return map[uuid.UUID]api.ProcessStatus{
		p.ID: m.ProcessStorage.GetState(p.ID).Status,
	}
}

// Get returns a process from its ID
func (m *Manager) Get(id uuid.UUID) (*Process, error) {
	p, ok := m.ProcessStorage.GetProcess(id)
	if !ok {
		return nil, fmt.Errorf("No such process: %s", id)
	}
	return m.loadProcess(p)
}

// ListActiveEmissions lists active emissions for a given target
func (m *Manager) ListActiveEmissions(target uuid.UUID) []storage.Emission {
	return m.ProcessStorage.ListActiveEmissions(target)
}

// ListProcessEmissions lists active emissions for a given process
func (m *Manager) ListProcessEmissions(process uuid.UUID) []storage.Emission {
	return m.ProcessStorage.ListProcessEmissions(process)
}

func (m *Manager) nextState(
	process storage.Process, state storage.ProcessState, logentry storage.ProcessLogEntry,
) (storage.ProcessState, error) {
	if state.NodeStates == nil {
		state.NodeStates = make(map[storage.ProcessNode]storage.ProcessNodeState)
	}
	graph, ok := m.GraphManager.GetGraph(process.GraphID)
	if !ok {
		// XXX Panic? I mean if this really occurs do we have any chance
		// to recover and not return an error to the caller?
		return storage.ProcessState{}, fmt.Errorf("Process linked to unknown graph: %s", process.GraphID)
	}

	p, err := m.loadProcess(process)
	if err != nil {
		panic(err)
	}
	engine := StateEngine{p, state, graph, m.Log}

	var node storage.ProcessNode

	if logentry.Emission != nil {
		node = storage.ProcessNode{
			ActorID: logentry.Emission.TargetActor,
			NodeID:  logentry.Emission.TargetInputRef.NodeID,
		}
		engine.HitInput(
			node,
			logentry.Emission.TargetInputRef.Input,
			logentry.EmissionPosition.Complete,
			false, // closed. For now no input close can come from the log entries
		)

	} else if logentry.Reception != nil {
		node = storage.ProcessNode{
			ActorID: logentry.Reception.SourceActor,
			NodeID:  logentry.Reception.SourceOutputRef.NodeID,
		}
		engine.HitOutput(
			node,
			logentry.Reception.SourceOutputRef.Output,
			logentry.ReceptionStatus,
		)
	} else if logentry.ProcessNodeEnd != nil {
		node = logentry.ProcessNodeEnd.Node
		engine.EndProcessNode(node, logentry.ProcessNodeEnd.Result)
	} else {
		panic("Invalid log entry")
	}

	return engine.GetState(), nil
}
