// Copyright 2016 Orus.io. All rights reserved.

package emitter

import (
	"runtime/debug"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// EmitterQueue is the name we use to queueSub to a queue with the emitter
	EmitterQueue = "emitter_queue"
)

// Options is used to create a Emitter
type Options struct {
	services.Options

	ReceptionTimeout time.Duration

	ActorStorage    storage.ActorStorage
	EnvelopeStorage storage.EnvelopeStorage
	ProcessManager  *process.Manager
}

// NewService creates a Emitter
func (o Options) NewService() *Service {
	if o.ReceptionTimeout == 0 {
		o.ReceptionTimeout = 2 * time.Second
	}

	return &Service{
		BaseService: services.NewBaseService("emitter", o.Options),
		o:           o,
	}
}

// Service is in charge of emitting envelopes to Actors
type Service struct {
	services.BaseService

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	s.Log.Noticef("Startup")

	if err := s.SubscribeAll(
		services.QueueSubscribe(api.EmitterSignalSubjet, EmitterQueue, s.handleSignal),
		services.QueueSubscribe(api.CoreActorEventsSubject, EmitterQueue, s.handleActorEvent),
		services.QueueSubscribe(api.CoreProcessEventsSubject, EmitterQueue, s.handleProcessEvent),
	); err != nil {
		s.Shutdown()
		return err
	}

	// if e.isAlone(), reschedule pending emissions

	return nil
}

func (s *Service) handleActorEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "up" {
		// if "up", reschedule emissions
		s.Log.Noticef("%s is back online", event.Actor.ID)
		emissions := s.o.ProcessManager.ListActiveEmissions(event.Actor.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmitterSignalSubjet, api.EmitterSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

func (s *Service) handleProcessEvent(subject string, event api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	if event.Name == "status" && event.Process.Status == api.ProcessRunning {
		s.Log.Noticef("Process %s is now running")

		emissions := s.o.ProcessManager.ListProcessEmissions(event.Process.ID)

		for _, pending := range emissions {
			s.Log.Noticef("Rescheduling %s to %s", pending.EnvelopeID, pending.TargetActor)
			s.Conn.Publish(api.EmitterSignalSubjet, api.EmitterSignal{
				ProcessID:  pending.ProcessID,
				EnvelopeID: pending.EnvelopeID,
				Target: api.EnvelopeTarget{
					ActorID: pending.TargetActor,
					NodeID:  pending.TargetInputRef.NodeID,
					Input:   pending.TargetInputRef.Input,
				},
			})
		}
	}
}

// emit as much of the env as possible and wait for a ack.
// if ok, touch the actor. If not, mark it invisibls.
// when the actor becomes visible again, the emission will
// be rescheduled
func (s *Service) handleSignal(subject string, signal api.EmitterSignal) {
	s.TaskAdd()
	process, err := s.o.ProcessManager.Get(signal.ProcessID)
	if err != nil {
		s.Log.Errorf("Received a wrong processID: %s", err)
		s.TaskDone()
		return
	}
	go s.emit(process, signal.EnvelopeID, signal.Target)
}

func (s *Service) emit(process *process.Process, envelopeID uuid.UUID, target api.EnvelopeTarget) {
	s.Log.Noticef("Sending to %s", target)
	defer s.SafeTaskDone()

	defer func() {
		if r := recover(); r != nil {
			s.Log.Noticef("Error in emit: %s\n%s", r, debug.Stack())
		}
	}()

	status := process.GetStatus()
	if status != api.ProcessRunning {
		s.Log.Noticef("Process not running (current state: %s), skip emission", status)
		return
	}

	emitPosition, err := process.GetEmitPosition(envelopeID, target)
	if err != nil {
		s.Log.Errorf("Cannot get emit position: %s", err)
		return
	}
	defer emitPosition.Unlock()

	pos, err := emitPosition.Get()
	if err != nil {
		s.Log.Errorf("Cannot read emit position: %s", err)
		return
	}
	newenv, newpos, err := s.o.EnvelopeStorage.ReadEnvelope(envelopeID, pos)
	if err != nil {
		s.Log.Errorf("Error reading envelope for transfer: %s", err)
		return
	}
	if newpos.Equals(pos) {
		s.Log.Debugf("No new data to emit")
		return
	}

	newenv.Context = api.EnvelopeContext{
		ProcessID: process.ID,
		NodeID:    target.NodeID,
	}

	actor, ok := s.o.ActorStorage.Get(target.ActorID)
	if !ok {
		s.Log.Errorf("Unknown actor")
		return
	}
	var ack api.EnvelopeAck
	outsubject := api.ActorInputSubject(target.ActorID, target.Input)
	// For the non-first fragments of envelope, publish to a dedicated subject
	if !pos.Start {
		outsubject += "." + envelopeID.String()
	}
	if actor.AccountID.IsZero() {
		err = s.Conn.Request(outsubject, newenv, &ack, s.o.ReceptionTimeout)
	} else {
		err = s.Conn.RequestFor(actor.AccountID.String(), outsubject, newenv, &ack, s.o.ReceptionTimeout)
	}
	if err == nats.ErrTimeout {
		s.Log.Errorf("Actor did not ack the envelops. Mark it unresponsivs.")
		if err := s.o.ActorStorage.MarkUnresponsive(target.ActorID); err != nil {
			s.Log.Errorf("Cannot mark the actor unresponsive: %s", err)
		}
		return
	} else if err != nil {
		s.Log.Errorf(
			"Error transmitting envelope to %s on %s: %s",
			target, outsubject, err)
		return
	}
	if ack.ReceptionStatus == api.ReceptionError {
		s.Log.Errorf("Actor returned an error: %s", ack.Reason)
	} else {
		s.Log.Noticef("Published to %s", outsubject)
		if err := emitPosition.Set(newpos); err != nil {
			s.Log.Errorf("Cannot set emit position: %s", err)
			return
		}
	}
}
