// Copyright 2016 Orus.io. All rights reserved.

package processstate

import (
	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/process"
	"bitbucket.org/orus-io/xbus/services"
	"bitbucket.org/orus-io/xbus/storage"
)

const (
	// ServiceName name of the service
	ServiceName = "processstate"

	// QueueName queue name for queuesubscribes
	QueueName = "xbus.services." + ServiceName
)

// Options is used to create a 'processstate' service
type Options struct {
	services.Options

	LogStorage     storage.LogStorage
	ProcessManager *process.Manager
}

// NewService creates a Service
func (o Options) NewService() *Service {
	return &Service{
		BaseService: services.NewBaseService(ServiceName, o.Options),
		o:           o,
	}
}

// Service is in charge of emitting process status events to various subjects
type Service struct {
	services.BaseService

	o Options
}

// Startup starts the service
func (s *Service) Startup() error {
	if err := s.SubscribeAll(
		services.QueueSubscribe(api.CoreProcessEventsSubject, QueueName, s.handleProcessEvent),
	); err != nil {
		return err
	}
	return nil
}

func (s *Service) handleProcessEvent(subject string, evt api.CoreEvent) {
	s.TaskAdd()
	defer s.SafeTaskDone()

	actor := api.Actor{ID: evt.Process.TriggerEmitterID}
	statusList := s.o.ProcessManager.GetEmitterEnvelopeStatus(
		actor, evt.Process.TriggerEnvelopeID,
	)
	if len(statusList) == 0 {
		panic("No process for trigger... which is weird if not impossible. Please report")
	}
	if len(statusList) != 1 {
		panic("Not implemented")
	}
	var combinedStatus api.ProcessStatus
	for _, status := range statusList {
		combinedStatus = status
	}
	var errors = make(map[uuid.UUID][]api.LogMessage)
	for pID := range statusList {
		for _, entry := range s.o.LogStorage.GetProcessMessages(
			pID, api.LogError,
		) {
			s.Log.Noticef("%+v", entry)
			errors[entry.ProcessID] = append(errors[entry.ProcessID], entry.Message)
		}
	}
	subj := api.EmitterEnvelopeStateSubject(
		evt.Process.TriggerEmitterID, evt.Process.TriggerEnvelopeID)
	if err := s.Conn.Publish(
		subj,
		api.EmitterEnvelopeState{
			ProcessStatusList: statusList,
			Status:            combinedStatus,
			Errors:            errors,
		},
	); err != nil {
		s.Log.Errorf("Error publishing to %s: %s", subj, err)
	}
}
