package services

import (
	"fmt"
	"runtime/debug"
	"sync"

	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// Options is the common options of xbus services
type Options struct {
	Conn   *nats.EncodedConn
	Logger xbus.Logger
}

// Service is the common interface of all the services
type Service interface {
	// GetName returns the service name
	GetName() string

	// Startup will do what is necessary for the service to be up and running
	Startup() error

	// Shutdown should stop the service and release any resource allocated by
	// Startup
	Shutdown() error
}

// BaseService is a base struct for implementing services.
type BaseService struct {
	Name string

	Log xbus.Logger

	Conn *tools.Conn
	subs []*nats.Subscription

	Tasks sync.WaitGroup
}

// NewBaseService initialize a BaseService
func NewBaseService(name string, o Options) BaseService {
	return BaseService{
		Name: name,
		Log: xbus.NewPrefixedLogger(
			fmt.Sprintf("[%s] ", name),
			o.Logger),
		Conn: &tools.Conn{EncodedConn: o.Conn},
	}
}

// ping
// subscribe to xbus.ping.services.<name>

// isAlone
// send a ping and check how many instances reply.
// if only one (us), return true

// GetName returns the service nams
func (s *BaseService) GetName() string {
	return s.Name
}

// Shutdown is a default implementation that calls UnsubscribeAll()
func (s *BaseService) Shutdown() error {
	err := s.UnsubscribeAll()
	s.Tasks.Wait()
	return err
}

// AddSubscription save a subscription to be unsubscribed with UnsubscribeAll
func (s *BaseService) AddSubscription(sub *nats.Subscription) {
	s.subs = append(s.subs, sub)
}

// SubscriptionDef is a definition of a nats subscription
type SubscriptionDef struct {
	Subject string
	Queue   string
	Handler nats.Handler
}

// Subscribe returns a SubscriptionDef for simple subscription
func Subscribe(subject string, handler nats.Handler) SubscriptionDef {
	return SubscriptionDef{
		Subject: subject,
		Handler: handler,
	}
}

// QueueSubscribe returns a SubscriptionDef for queue subscription
func QueueSubscribe(subject, queue string, handler nats.Handler) SubscriptionDef {
	return SubscriptionDef{
		Subject: subject,
		Queue:   queue,
		Handler: handler,
	}
}

// SubscribeAll do multiple subscriptions at once and returns
// immediately on the first failing subscription.
func (s *BaseService) SubscribeAll(subdefs ...SubscriptionDef) error {
	for _, subdef := range subdefs {
		sub, err := s.Conn.QueueSubscribe(subdef.Subject, subdef.Queue, subdef.Handler)
		if err != nil {
			return err
		}
		s.AddSubscription(sub)
	}
	return nil
}

// UnsubscribeAll unsubscribe and forget all the subscriptions
func (s *BaseService) UnsubscribeAll() error {
	var lastError error
	for _, sub := range s.subs {
		err := sub.Unsubscribe()
		if err != nil {
			s.Log.Errorf("Error unsubscribing from %s: %s", sub.Subject, err)
			lastError = err
		}
	}
	return lastError
}

// SendReply sends a reply to the caller AND calls TaskDone(). TaskAdd() _must_ be
// called prior to SendReply
func (s *BaseService) SendReply(subject string, reply interface{}) {
	if r := recover(); r != nil {
		s.Log.Errorf("Caught error: %s\n%s", r, debug.Stack())
		reply = api.Response{
			Status:        api.StatusServerError,
			StatusMessage: fmt.Sprintf("%s", r),
		}
	}
	if err := s.Conn.Publish(subject, reply); err != nil {
		s.Log.Errorf("Error replying: %s", err)
	}
	s.TaskDone()
}

// HandlePanics handle any panics to avoid a server crash.
// panics will be logged with a stack trace and a message encouraging the
// user to report the issue
func (s *BaseService) HandlePanics() {
	if r := recover(); r != nil {
		s.Log.Errorf("Caught error: %s\n%s", r, debug.Stack())
	}
}

// TaskAdd signal that a task just started.
// It will block service shutdown until TaskDone is called.
func (s *BaseService) TaskAdd() {
	s.Tasks.Add(1)
}

// TaskDone signal the end of a task, allowing the service to shutdown
// if no other task is running.
func (s *BaseService) TaskDone() {
	s.Tasks.Done()
}

// SafeTaskDone is a version of TaskDone that recover panics and log them.
func (s *BaseService) SafeTaskDone() {
	if r := recover(); r != nil {
		s.Log.Errorf("Caught error: %s\n%s", r, debug.Stack())
	}
	s.Tasks.Done()
}
