package storage

import (
	"fmt"
	"strings"
)

func parseNodeIO(str string) (nodeID string, ioName string, err error) {
	for str[0] == ' ' {
		str = str[1:]
	}
	for str[len(str)-1] == ' ' {
		str = str[:len(str)-1]
	}
	splitted := strings.Split(str, ".")
	if len(splitted) > 2 {
		err = fmt.Errorf("Invalid NodeInputRef '%s': expects at most one '.'", str)
		return
	}
	nodeID = splitted[0]

	if len(splitted) == 2 {
		ioName = splitted[1]
	} else {
		ioName = "default"
	}
	return
}

// NodeInputRefFromString loads a NodeInputRef
func NodeInputRefFromString(str string) (ref NodeInputRef, err error) {
	ref.NodeID, ref.Input, err = parseNodeIO(str)
	return
}

// String serialize a NodeInputRef
func (r NodeInputRef) String() string {
	return r.NodeID + "." + r.Input
}

// MarshalYAML marshals NodeInputRef to string
func (r NodeInputRef) MarshalYAML() (interface{}, error) {
	return r.String(), nil
}

// UnmarshalYAML unmarshals NodeInputRef from string
func (r *NodeInputRef) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var str string
	var err error
	if err = unmarshal(&str); err != nil {
		return err
	}
	r.NodeID, r.Input, err = parseNodeIO(str)
	return err
}

// NodeOutputRefFromString loads a NodeInputRef
func NodeOutputRefFromString(str string) (ref NodeOutputRef, err error) {
	ref.NodeID, ref.Output, err = parseNodeIO(str)
	return
}

// String serialize a NodeOutputRef
func (r NodeOutputRef) String() string {
	return r.NodeID + "." + r.Output
}

// MarshalYAML marshals NodeOutputRef to string
func (r NodeOutputRef) MarshalYAML() (interface{}, error) {
	return r.String(), nil
}

// UnmarshalYAML unmarshals NodeOutputRef from string
func (r *NodeOutputRef) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var str string
	var err error
	if err = unmarshal(&str); err != nil {
		return err
	}
	r.NodeID, r.Output, err = parseNodeIO(str)
	return err
}

// EdgeFromString unserialize a Edge
func EdgeFromString(str string) (Edge, error) {
	sp := strings.Split(str, "->")
	if len(sp) != 2 {
		return Edge{}, fmt.Errorf("Invalid edge: %s", str)
	}
	from, err := NodeOutputRefFromString(sp[0])
	if err != nil {
		return Edge{}, fmt.Errorf("Syntax error: output '%s'", str)
	}
	to, err := NodeInputRefFromString(sp[1])
	if err != nil {
		return Edge{}, fmt.Errorf("Syntax error: input: '%s'", str)
	}
	return Edge{from, to}, nil
}

// String() serialize a Edge
func (e Edge) String() string {
	return e.From.String() + "->" + e.To.String()
}

// MarshalYAML marshals Edge to string
func (e Edge) MarshalYAML() (interface{}, error) {
	return e.String(), nil
}

// UnmarshalYAML unmarshals Edge from string
func (e *Edge) UnmarshalYAML(unmarshal func(interface{}) error) error {
	var (
		str string
		err error
	)
	if err = unmarshal(&str); err != nil {
		return err
	}
	*e, err = EdgeFromString(str)
	return err
}
