// Copyright 2016 Orus.io. All rights reserved.

package sql

import (
	"encoding/json"
	"fmt"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/xbus/storage"
	"bitbucket.org/orus-io/xbus/storage/sql/model"
)

// NewProcessStorage initialize a new ProcessStorage
func NewProcessStorage(model *model.Model, db *yago.DB, logger xbus.Logger) *ProcessStorage {
	return &ProcessStorage{
		Base: Base{
			model:  model,
			db:     db,
			logger: logger,
		},
		nextState:       storage.UndefinedNextStateCallback,
		onStatusChanged: storage.NoOpProcessEventStatusHandler,
	}
}

// ProcessStorage implements storage.ProcessStorage
type ProcessStorage struct {
	Base
	nextState       storage.NextStateCallback
	onStatusChanged storage.ProcessEventStatusHandler
}

// SetEventHandlers sets the event handlers
func (s *ProcessStorage) SetEventHandlers(onStatusChanged storage.ProcessEventStatusHandler) {
	if onStatusChanged == nil {
		s.onStatusChanged = storage.NoOpProcessEventStatusHandler
	} else {
		s.onStatusChanged = onStatusChanged
	}
}

func dbProcessToStorage(dbProcess model.Process) storage.Process {
	return storage.Process{
		ID:                dbProcess.ID,
		TriggerEmitterID:  dbProcess.TriggerEmitterID,
		TriggerEnvelopeID: dbProcess.TriggerEnvelopeID,
		GraphID:           dbProcess.GraphID,
		SourceOutput: storage.NodeOutputRef{
			NodeID: dbProcess.SourceNodeID,
			Output: dbProcess.SourceSlotID,
		},
	}
}

func storageProcessToDB(process storage.Process) model.Process {
	return model.Process{
		Base:              model.Base{ID: process.ID},
		TriggerEmitterID:  process.TriggerEmitterID,
		TriggerEnvelopeID: process.TriggerEnvelopeID,
		GraphID:           process.GraphID,
		SourceNodeID:      process.SourceOutput.NodeID,
		SourceSlotID:      process.SourceOutput.Output,
	}
}

func (s *ProcessStorage) loadProcess(tx *yago.Tx, id uuid.UUID) (storage.Process, bool) {
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Get(&dbProcess, id); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// GetProcess returns a process data
func (s *ProcessStorage) GetProcess(id uuid.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	return s.loadProcess(tx.Tx, id)
}

// GetProcessByTrigger returns a process matching a trigger or an empty value if not found
// TODO in the future we should support multiple process per trigger.
func (s *ProcessStorage) GetProcessByTrigger(emitter api.Actor, envelopeID uuid.UUID) (storage.Process, bool) {
	tx := s.MustBegin()
	defer tx.Rollback()
	var dbProcess model.Process
	if err := tx.Query(s.model.Process).Where(
		s.model.Process.TriggerEmitterID.Eq(emitter.ID),
		s.model.Process.TriggerEnvelopeID.Eq(envelopeID),
	).One(&dbProcess); err != nil {
		if err == yago.ErrRecordNotFound {
			return storage.Process{}, false
		}
		must(err)
	}
	return dbProcessToStorage(dbProcess), true
}

// Query lookup for processes
func (s *ProcessStorage) Query(includeTerminated bool) []storage.Process {
	var dbProcessList []model.Process
	tx := s.MustBegin()
	defer tx.Rollback()
	q := tx.Query(s.model.Process)
	if !includeTerminated {
		q = q.LeftJoin(s.model.ProcessState).Filter(
			s.model.ProcessState.Status.Eq(api.ProcessRunning),
		)
	}
	must(
		q.All(&dbProcessList))
	var pList []storage.Process
	for _, dbP := range dbProcessList {
		pList = append(pList, dbProcessToStorage(dbP))
	}
	return pList
}

// CreateProcess saves a process
func (s *ProcessStorage) CreateProcess(p storage.Process) (uuid.UUID, error) {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	p.ID = uuid.UUID{}

	dbProcess := storageProcessToDB(p)

	must(tx.Insert(&dbProcess))

	tx.MustCommit()

	return dbProcess.ID, nil
}

// UpdateReceiveStatus updates a reception status
func (s *ProcessStorage) UpdateReceiveStatus(reception storage.Reception, status storage.ReceptionStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogReception(tx, reception.ProcessID, reception, status)
	must(err)

	return s.updateStateAndCommit(
		tx, reception.ProcessID, []storage.ProcessLogEntry{logentry})
}

// ProcessNodeEnd signals the end of a ProcessNode.
func (s *ProcessStorage) ProcessNodeEnd(
	processID uuid.UUID, node storage.ProcessNode,
	envelopeID uuid.UUID, result api.ActorProcessingResultType,
) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	logentry, err := s.processLogProcessNodeEnd(
		tx, processID,
		storage.ProcessNodeEnd{
			Node:       node,
			EnvelopeID: envelopeID,
			Result:     result,
		})
	must(err)

	return s.updateStateAndCommit(
		tx, processID, []storage.ProcessLogEntry{logentry})
}

// GetEmitPosition returns an exclusive access to an emit position
func (s *ProcessStorage) GetEmitPosition(emission storage.Emission) (storage.EmitPosition, error) {
	var ownTx = true
	tx := s.MustBegin()
	defer func() {
		if ownTx {
			tx.Rollback()
		}
	}()
	// TODO Use an exclusive transaction
	// on sqlite: BEGIN EXCLUSIVE TRANSACTION
	// on PG: SET TRANSACTION SERIALIZABLE
	if false && s.db.Engine.Dialect().Driver() == "postgres" {
		_, err := tx.GetEngine().Exec(model.RawSQL("SET TRANSACTION ISOLATION LEVEL SERIALIZABLE"))
		if err != nil {
			panic(err)
		}
	}
	// They load the value from the db
	var dbEmission model.Emission
	var pos = storage.EnvelopeStartPos()
	if err := tx.Query(s.model.Emission).
		Where(
			s.model.Emission.ProcessID.Eq(emission.ProcessID),
			s.model.Emission.EnvelopeID.Eq(emission.EnvelopeID),
			s.model.Emission.TargetActor.Eq(emission.TargetActor),
			s.model.Emission.TargetNodeID.Eq(emission.TargetInputRef.NodeID),
			s.model.Emission.TargetInput.Eq(emission.TargetInputRef.Input),
		).
		ForUpdate(s.model.Emission).
		One(&dbEmission); err != nil {
		if err == yago.ErrRecordNotFound {
			dbEmission.ProcessID = emission.ProcessID
			dbEmission.EnvelopeID = emission.EnvelopeID
			dbEmission.TargetActor = emission.TargetActor
			dbEmission.TargetNodeID = emission.TargetInputRef.NodeID
			dbEmission.TargetInput = emission.TargetInputRef.Input
			dbEmission.Start = true
			dbEmission.EventPositions = "{}"
			err = tx.Insert(&dbEmission)
		}
		if err != nil {
			panic(err)
		}
	}
	pos.Start = dbEmission.Start
	pos.Complete = dbEmission.Complete
	if err := json.Unmarshal([]byte(dbEmission.EventPositions), &pos.Events); err != nil {
		panic(err)
	}
	ownTx = false
	return &EmitPosition{
		tx:             tx,
		emission:       dbEmission,
		pos:            pos,
		processStorage: s,
	}, nil
}

func (s *ProcessStorage) list(clauses ...qb.Clause) []storage.Emission {
	var dbEmissionList []model.Emission
	tx := s.MustBegin()
	defer tx.Rollback()
	must(
		tx.Query(s.model.Emission).
			Where(clauses...).
			All(&dbEmissionList))
	var emissionList []storage.Emission
	for _, e := range dbEmissionList {
		emissionList = append(emissionList, storage.Emission{
			ProcessID:   e.ProcessID,
			EnvelopeID:  e.EnvelopeID,
			TargetActor: e.TargetActor,
			TargetInputRef: storage.NodeInputRef{
				NodeID: e.TargetNodeID,
				Input:  e.TargetInput,
			},
		})
	}
	return emissionList
}

// ListActiveEmissions returns the active emissions
func (s *ProcessStorage) ListActiveEmissions(targetActor uuid.UUID) []storage.Emission {
	return s.list(
		s.model.Emission.Complete.Eq(false),
		s.model.Emission.TargetActor.Eq(targetActor),
	)
}

// ListProcessEmissions returns the active emissions
func (s *ProcessStorage) ListProcessEmissions(process uuid.UUID) []storage.Emission {
	return s.list(
		s.model.Emission.Complete.Eq(false),
		s.model.Emission.ProcessID.Eq(process),
	)
}

// GetState returns an instant copy of the state of a process
func (s *ProcessStorage) GetState(processID uuid.UUID) storage.ProcessState {
	tx := s.MustBegin()
	defer tx.Rollback()
	state, err := s.loadState(tx.Tx, processID, false)
	must(err)
	return state
}

// SetStatus changes the current status. Only Runnning->Paused and Paused->Running
// are allowed at this point
func (s *ProcessStorage) SetStatus(processID uuid.UUID, status api.ProcessStatus) error {
	tx := s.MustBegin()
	defer tx.RollbackIfOpened()

	process, ok := s.loadProcess(tx.Tx, processID)
	if !ok {
		return fmt.Errorf("No such process: %s", processID)
	}

	state, err := s.loadState(tx.Tx, processID, false)
	must(err)

	if !(status == api.ProcessPaused && state.Status == api.ProcessRunning ||
		status == api.ProcessRunning && state.Status == api.ProcessPaused) {
		return fmt.Errorf("Invalid state transition: %s -> %s", state.Status, status)
	}

	old := state.Status
	state.Status = status

	err = s.saveState(tx.Tx, processID, state)
	must(err)

	tx.MustCommit()

	s.onStatusChanged(process, old, state.Status)

	return nil
}

// SetNextStateCallback ...
func (s *ProcessStorage) SetNextStateCallback(cb storage.NextStateCallback) {
	if cb == nil {
		s.nextState = storage.UndefinedNextStateCallback
	} else {
		s.nextState = cb
	}
}

func (s *ProcessStorage) loadState(tx *yago.Tx, processID uuid.UUID, forUpdate bool) (pState storage.ProcessState, err error) {
	var dbProcessState model.ProcessState

	q := tx.Query(s.model.ProcessState)
	if forUpdate {
		q = q.ForUpdate()
	}
	if err = q.Get(&dbProcessState, processID); err != nil && err != yago.ErrRecordNotFound {
		return pState, err
	}

	if dbProcessState.NodeStates != "" {
		pState.Status = dbProcessState.Status
		if err = json.Unmarshal(
			[]byte(dbProcessState.NodeStates),
			&pState.NodeStates); err != nil {
			return
		}
	}
	return pState, nil
}

func (s *ProcessStorage) saveState(tx *yago.Tx, processID uuid.UUID, pState storage.ProcessState) error {
	b, err := json.Marshal(pState.NodeStates)
	if err != nil {
		return err
	}
	dbProcessState := model.ProcessState{
		ProcessID:  processID,
		Status:     pState.Status,
		NodeStates: string(b),
	}
	exists, err := tx.Query(s.model.ProcessState).
		Where(s.model.ProcessState.ProcessID.Eq(processID)).
		Exists()
	if err != nil {
		return err
	}
	if !exists {
		return tx.Insert(&dbProcessState)
	}
	return tx.Update(&dbProcessState)
}

func (s *ProcessStorage) processLogEmission(tx *TxHelper, processID uuid.UUID, emission storage.Emission, pos storage.EnvelopeReadPos) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		Emission:         &emission,
		EmissionPosition: &pos,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogReception(tx *TxHelper, processID uuid.UUID, reception storage.Reception, status storage.ReceptionStatus) (storage.ProcessLogEntry, error) {
	logentry := storage.ProcessLogEntry{
		Reception:       &reception,
		ReceptionStatus: status,
	}
	return s.processLog(tx, processID, logentry)
}

func (s *ProcessStorage) processLogProcessNodeEnd(
	tx *TxHelper, processID uuid.UUID, nodeEnd storage.ProcessNodeEnd,
) (storage.ProcessLogEntry, error) {
	return s.processLog(
		tx, processID,
		storage.ProcessLogEntry{
			ProcessNodeEnd: &nodeEnd,
		})
}

func (s *ProcessStorage) processLog(tx *TxHelper, processID uuid.UUID, logentry storage.ProcessLogEntry) (storage.ProcessLogEntry, error) {
	logentry.Timestamp = time.Now()
	b, err := json.Marshal(logentry)
	if err != nil {
		return storage.ProcessLogEntry{}, err
	}
	dbLogEntry := model.ProcessLogEntry{
		ProcessID: processID,
		Timestamp: logentry.Timestamp,
		Data:      string(b),
	}
	if err := tx.Insert(&dbLogEntry); err != nil {
		return storage.ProcessLogEntry{}, err
	}
	return logentry, nil
}

func (s *ProcessStorage) updateStateAndCommit(tx *TxHelper, processID uuid.UUID, logentries []storage.ProcessLogEntry) error {
	process, ok := s.GetProcess(processID)
	if !ok {
		return fmt.Errorf("No such process:  %s", processID)
	}
	pState, err := s.loadState(tx.Tx, processID, true)
	must(err)
	statusList := []api.ProcessStatus{pState.Status}
	for _, logentry := range logentries {
		pState, err = s.nextState(process, pState, logentry)
		must(err)

		if pState.Status != statusList[len(statusList)-1] {
			statusList = append(statusList, pState.Status)
		}
	}
	must(
		s.saveState(tx.Tx, processID, pState))

	tx.MustCommit()

	// Emit a StatusChanged for each transition
	for i := range statusList[1:] {
		s.onStatusChanged(process, statusList[i], statusList[i+1])
	}
	return err
}

// EmitPosition implements storage.EmitPosition
type EmitPosition struct {
	tx             *TxHelper
	emission       model.Emission
	pos            storage.EnvelopeReadPos
	processStorage *ProcessStorage
	logentries     []storage.ProcessLogEntry
	lastError      error
}

// Get returns the position
func (p *EmitPosition) Get() (storage.EnvelopeReadPos, error) {
	return p.pos, p.lastError
}

// Set updates the position
func (p *EmitPosition) Set(pos storage.EnvelopeReadPos) error {
	if p.lastError != nil {
		return p.lastError
	}
	p.pos = pos
	p.emission.Start = pos.Start
	p.emission.Complete = pos.Complete
	b, err := json.Marshal(pos.Events)
	p.lastError = err
	must(err)
	p.emission.EventPositions = string(b)

	p.lastError = p.tx.Update(
		&p.emission,
		model.EmissionStart,
		model.EmissionComplete,
		model.EmissionEventPositions)
	must(p.lastError)
	logentry, err := p.processStorage.processLogEmission(
		p.tx,
		p.emission.ProcessID,
		storage.Emission{
			ProcessID:   p.emission.ProcessID,
			EnvelopeID:  p.emission.EnvelopeID,
			TargetActor: p.emission.TargetActor,
			TargetInputRef: storage.NodeInputRef{
				NodeID: p.emission.TargetNodeID,
				Input:  p.emission.TargetInput,
			},
		},
		pos)
	p.lastError = err
	must(err)
	p.logentries = append(p.logentries, logentry)
	return nil
}

// Unlock releases the lock on the position
func (p *EmitPosition) Unlock() error {
	if p.lastError != nil {
		p.tx.Rollback()
		return nil
	}
	return p.processStorage.updateStateAndCommit(
		p.tx, p.emission.ProcessID, p.logentries)
}
