// Copyright 2016 Orus.io. All rights reserved.

package mocks

import (
	"fmt"

	"github.com/m4rw3r/uuid"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus"
	"bitbucket.org/orus-io/go-xbus/tools"
	"bitbucket.org/orus-io/xbus/storage"
)

// NewEnvelopeStorage init an empty EnvelopeStorage suitable for tests
func NewEnvelopeStorage(logger xbus.Logger) *EnvelopeStorage {
	return &EnvelopeStorage{
		envelopes:          make(map[uuid.UUID]api.Envelope),
		envelopeValidators: make(map[uuid.UUID]*tools.EnvelopeValidator),
		logger:             logger,
	}
}

// EnvelopeStorage implements EnvelopeStorage on top of a SQL database
type EnvelopeStorage struct {
	envelopes          map[uuid.UUID]api.Envelope
	envelopeValidators map[uuid.UUID]*tools.EnvelopeValidator
	logger             xbus.Logger
}

// GetEventTypes returns an envelop event types
func (s *EnvelopeStorage) GetEventTypes(id uuid.UUID) ([]string, bool) {
	envelope, ok := s.envelopes[id]
	if !ok {
		return nil, false
	}
	var eventTypes []string
	for _, event := range envelope.Events {
		eventTypes = append(eventTypes, event.Type)
	}
	return eventTypes, true
}

// StoreEnvelope store a part of an envelope and returns its status
// in the storage
// TODO handle events arriving in the wrong order.
func (s *EnvelopeStorage) StoreEnvelope(e api.Envelope) (storage.EnvelopeStatus, error) {
	var err error
	envelope, ok := s.envelopes[e.ID]
	if !ok {
		envelope = e
		s.envelopeValidators[e.ID], err = tools.NewEnvelopeValidator(envelope)
		if err != nil {
			return storage.EnvelopeStatus{}, err
		}
	} else {
		err = s.envelopeValidators[e.ID].Add(e)
		if err != nil {
			return storage.EnvelopeStatus{}, err
		}
		if len(e.EventIDs) != 0 {
			envelope.EventIDs = e.EventIDs
		}
		for _, newEvent := range e.Events {
			for i, event := range envelope.Events {
				if newEvent.ID == event.ID {
					event.Items = append(event.Items, newEvent.Items...)
					if newEvent.Type != "" {
						if event.Type != "" && event.Type != newEvent.Type {
							return storage.EnvelopeStatus{}, fmt.Errorf("Inconsistent types on event %s", event.ID)
						}
						if event.Type == "" {
							event.Type = newEvent.Type
						}
					}
					if newEvent.Checksum != 0 {
						event.Checksum = newEvent.Checksum
					}
					if newEvent.ItemCount != 0 {
						event.ItemCount = newEvent.ItemCount
					}
					envelope.Events[i] = event
				} else {
					envelope.Events = append(envelope.Events, event)
				}
			}
		}
	}
	s.envelopes[e.ID] = envelope

	status := storage.EnvelopeStatus{
		ID:     envelope.ID,
		Status: "complete",
	}

	v := s.envelopeValidators[envelope.ID]

	expectedIds := envelope.EventIDs
	var unknownEventTypes = false

	for _, event := range envelope.Events {
		for i, id := range expectedIds {
			if event.ID == id {
				expectedIds = append(expectedIds[:i], expectedIds[i+1:]...)
				break
			}
		}
		if event.Type == "" {
			unknownEventTypes = true
		}
		ev := v.Events[event.ID]
		var evStatus = storage.EventStatus{
			ID:             event.ID,
			Type:           event.Type,
			AvailableItems: len(event.Items),
			TotalItems:     event.ItemCount,
		}
		if !ev.Complete {
			evStatus.Status = "receiving"
			if status.Status == "complete" {
				status.Status = "receiving"
			}
		} else {
			if ev.Valid {
				evStatus.Status = "complete"
			} else {
				evStatus.Status = "error"
				evStatus.Err = ev.Error
				status.Status = "error"
			}
		}
		status.Events = append(status.Events, evStatus)
	}

	status.TypesKnown = len(envelope.EventIDs) != 0 && len(expectedIds) == 0 && !unknownEventTypes

	return status, nil
}

// ReadEnvelope reads as much as it can of an envelope, starting at 'position'
// The returned envelope should be suitable for encrypted nats transfert, ie
// its total jsonified size should be under api.EnvelopeMaxSize
func (s *EnvelopeStorage) ReadEnvelope(id uuid.UUID, position storage.EnvelopeReadPos) (api.Envelope, storage.EnvelopeReadPos, error) {
	srcEnv, ok := s.envelopes[id]
	if !ok {
		return api.Envelope{}, position, fmt.Errorf("No such envelope: %s", id)
	}
	// TODO verify that all the event types are known. Return an error
	if position.Start {
		for _, e := range srcEnv.Events {
			position.Events[e.ID] = storage.EventReadPos{}
		}
		position.Start = false
	}
	tgtEnv := api.Envelope{
		ID:       srcEnv.ID,
		EventIDs: srcEnv.EventIDs,
	}

	// TODO in case we already know the envelope is small enough, return
	// it complete at once

	// approx size of the jsonified envelope before any event is added
	var estimateSize = 36 + len(srcEnv.EventIDs)*3

	position = position.Clone()

	tgtEvents := map[uuid.UUID]api.Event{}

	for _, ev := range srcEnv.Events {
		eventPos := position.Events[ev.ID]
		if eventPos.ItemCount < ev.ItemCount {
			eventPos.Index++
			tgtEvents[ev.ID] = api.Event{
				ID:        ev.ID,
				Type:      ev.Type,
				Index:     eventPos.Index,
				ItemCount: ev.ItemCount,
				Checksum:  ev.Checksum,
			}
			estimateSize += 36 + len(ev.Type) + 8
		}
		position.Events[ev.ID] = eventPos
	}

	for estimateSize < api.EnvelopeMaxSize {
		before := estimateSize
		for _, ev := range srcEnv.Events {
			pos := position.Events[ev.ID]
			if pos.ItemCount == ev.ItemCount {
				continue
			}
			tgt := tgtEvents[ev.ID]
			item := ev.Items[pos.ItemCount]
			tgt.Items = append(tgt.Items, item)
			pos.ItemCount++

			estimateSize += int(float64(len(item)) * 0.1)

			position.Events[ev.ID] = pos
			tgtEvents[ev.ID] = tgt
			if estimateSize >= api.EnvelopeMaxSize {
				break
			}
		}
		if estimateSize == before {
			break
		}
	}

	complete := true
	for _, ev := range tgtEvents {
		tgtEnv.Events = append(tgtEnv.Events, ev)
		if ev.ItemCount != position.Events[ev.ID].ItemCount {
			complete = false
		}
	}

	if complete {
		tgtEnv.Last = true
		position.Complete = true
	}

	return tgtEnv, position, nil
}
