// Copyright 2016 Orus.io. All rights reserved.

package sqldb

import (
	"fmt"
	"os"
	"testing"

	// load postgres driver
	_ "github.com/lib/pq"
	// load sqlite driver
	_ "github.com/mattn/go-sqlite3"
	"github.com/orus-io/yago"
	"github.com/slicebit/qb"
	"github.com/tgulacsi/go-locking"

	"bitbucket.org/orus-io/xbus/storage/sql/model"
	"bitbucket.org/orus-io/xbus/testutils"
)

// CloseFunction is a function with no arg nor return values for closing a resource
type CloseFunction func()

// GetTestDB returns a model&db
func GetTestDB(t *testing.T) (*model.Model, *yago.DB, CloseFunction) {
	var (
		driver     = "sqlite3"
		droptables = false
	)

	dsn, ok := os.LookupEnv("XBUS_TEST_POSTGRES_DSN")
	if ok {
		driver = "postgres"
		droptables = true
	} else {
		dsn, ok = os.LookupEnv("XBUS_TEST_SQLITE_DSN")
		if !ok {
			os.Remove("test.db")
			dsn = "test.db"
		}
	}
	logFlags := qb.LDefault
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGQUERY"); v != "" {
		logFlags |= qb.LQuery
	}
	if v, _ := os.LookupEnv("XBUS_TEST_SQL_LOGBINDINGS"); v != "" {
		logFlags |= qb.LBindings
	}
	close := func() {}
	m, db, err := model.NewDB(driver, dsn, testutils.GetTestLogger(t), logFlags)
	if err != nil {
		t.Fatalf("Cannot initialize test database: %s", err)
		return nil, nil, nil
	}
	var allGood bool
	defer func() {
		if !allGood {
			close()
		}
	}()
	if driver == "postgres" {
		lock := locking.NewPortLock(1764)
		fmt.Println("Getting a lock")
		err := lock.Lock()
		if err != nil {
			t.Errorf("Cannot get lock: %s", err)
			return nil, nil, nil
		}
		fmt.Println("Got it")
		close = func() {
			err := lock.Unlock()
			if err != nil {
				t.Errorf("Error releasing lock: %s", err)
			}
			fmt.Println("Released it")
		}
	}
	if droptables {
		tables := db.Metadata.GetQbMetadata().Tables()
		for i := len(tables) - 1; i >= 0; i-- {
			table := tables[i]
			drop := table.Drop(db.Engine.Dialect())
			if _, err := db.Engine.DB().Exec(drop); err != nil {
				t.Log(err)
			}
		}
	}

	if err := db.Metadata.GetQbMetadata().CreateAll(db.Engine); err != nil {
		fmt.Println(err)
		t.Fatal(err)
	}
	allGood = true
	return m, db, close
}
