// Copyright 2016 Orus.io. All rights reserved.

package storage

import (
	"testing"

	"github.com/m4rw3r/uuid"
	"github.com/stretchr/testify/assert"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/xbus/storage"
)

// EnvelopeStorageFactory is any factory of EnvelopeStorage
type EnvelopeStorageFactory func() storage.EnvelopeStorage

// EnvelopeStorageTests is a test suite for EnvelopeStorage implementations
func EnvelopeStorageTests(t *testing.T, factory EnvelopeStorageFactory) {

	t.Run("Store/SingleEnvelope", func(t *testing.T) {
		envelopeStorage := factory()
		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, 1, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "complete", status.Events[0].Status)
		assert.Equal(t, "complete", status.Status)

		types, exists := envelopeStorage.GetEventTypes(env.ID)
		assert.True(t, exists)
		assert.Equal(t, []string{"xbus.test"}, types)
	})

	t.Run("Store/PairOfEnvelope", func(t *testing.T) {
		envelopeStorage := factory()

		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:    uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index: 1,
					Type:  "xbus.test",
					Items: []api.Item{[]byte("Data")},
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.True(t, status.TypesKnown)
		assert.Equal(t, 1, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "receiving", status.Events[0].Status)
		assert.Equal(t, "receiving", status.Status)

		env = api.Envelope{
			ID: env.ID,
			Events: []api.Event{
				api.Event{
					ID:        eventID,
					Index:     2,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 2,
					Checksum:  1307176161,
				},
			},
		}

		env.EventIDs = []uuid.UUID{eventID}

		status, err = envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, env.ID, status.ID)
		assert.Equal(t, 1, len(status.Events))
		if t.Failed() {
			return
		}
		assert.Equal(t, eventID, status.Events[0].ID)
		assert.Equal(t, 2, status.Events[0].AvailableItems)
		assert.Equal(t, nil, status.Events[0].Err)
		assert.Equal(t, "complete", status.Events[0].Status)
		assert.Equal(t, "complete", status.Status)
	})

	t.Run("Read/SingleEnvelope", func(t *testing.T) {
		envelopeStorage := factory()
		env := api.Envelope{
			ID: uuid.MustFromString("b74415ee-83c1-11e6-8f04-133b336287df"),
			Events: []api.Event{
				api.Event{
					ID:        uuid.MustFromString("c2e373fe-83c1-11e6-a098-c3d72ea43882"),
					Index:     1,
					Type:      "xbus.test",
					Items:     []api.Item{[]byte("Data")},
					ItemCount: 1,
					Checksum:  1859455373,
				},
			},
		}

		eventID := env.Events[0].ID
		env.EventIDs = []uuid.UUID{eventID}

		status, err := envelopeStorage.StoreEnvelope(env)
		assert.Nil(t, err)
		assert.Equal(t, "complete", status.Status)

		out, pos, err := envelopeStorage.ReadEnvelope(env.ID, storage.EnvelopeStartPos())
		assert.Nil(t, err)
		assert.True(t, pos.Complete)
		assert.Equal(t, env.ID, out.ID)
		assert.Equal(t, 1, len(out.Events))
		if t.Failed() {
			return
		}
		outEvent := out.Events[0]
		assert.Equal(t, eventID, outEvent.ID)
		assert.Equal(t, 1, outEvent.Index)
		assert.Equal(t, "xbus.test", outEvent.Type)
		assert.Equal(t, []api.Item{[]byte("Data")}, outEvent.Items)
		assert.Equal(t, 1, outEvent.ItemCount)
		assert.Equal(t, api.Checksum(1859455373), outEvent.Checksum)
	})
}
