// Copyright 2017 Orus.io. All rights reserved.

package xbus

import (
	"fmt"
	"strings"
	"time"

	"github.com/m4rw3r/uuid"
	"github.com/nats-io/go-nats"

	"bitbucket.org/orus-io/go-xbus/api"
	"bitbucket.org/orus-io/go-xbus/tools"
)

// EnvelopeHandler is a type of callbacks for when new envelope arrive on an input
type EnvelopeHandler func(string, *EnvelopeReader)

// newActor initialize an actor
func newActor(client *Client, options ActorOptions) *Actor {
	actor := Actor{
		ID:      options.ID,
		Name:    options.Name,
		Kind:    options.Kind,
		Options: options,
		Client:  client,
	}
	actor.Log = &ActorLogger{
		NewPrefixedLogger("[actor/"+options.Name+"] ", client.Log),
		&actor}
	return &actor
}

// Actor provides the common tools for all actors
type Actor struct {
	ID      uuid.UUID
	Name    string
	Kind    api.ActorKind
	Options ActorOptions
	Log     *ActorLogger
	Client  *Client
}

// GetStringSetting returns a string typed setting entry
func (a *Actor) GetStringSetting(name string) string {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return ""
	}
	sValue, ok := value.(string)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value: %s", value))
	}
	return sValue
}

// MustGetStringSetting returns a string setting or panic if undefined
func (a *Actor) MustGetStringSetting(name string) string {
	value := a.GetStringSetting(name)
	if value == "" {
		a.Client.Log.Fatalf("Missing setting: %s", name)
	}
	return value
}

// GetIntSettingD returns a int setting, or the defaultValue if not specified
func (a *Actor) GetIntSettingD(name string, defaultValue int) int {
	value, ok := a.Options.Service.Settings[name]
	if !ok {
		return defaultValue
	}
	iValue, ok := value.(int)
	if !ok {
		panic(fmt.Sprintf("Invalid configuration value '%s'. Expected an integer", value))
	}
	return iValue
}

// GetClient returns the underlying Client instance
func (a *Actor) GetClient() *Client {
	return a.Client
}

// ProcessingEnd calls the ActorProcessingEnd xbus api
func (a *Actor) ProcessingEnd(
	envelope *api.Envelope,
	result api.ActorProcessingResultType,
	messages []api.LogMessage,
) error {
	var rep api.ActorProcessingEndRep
	if err := a.Client.GetConn().Request(
		api.ActorProcessingEnd(a.ID),
		api.ActorProcessingEndReq{
			EnvelopeID: envelope.ID,
			Context:    envelope.Context,
			Result:     result,
			Messages:   messages,
		},
		&rep,
		10*time.Second,
	); err != nil {
		return err
	}
	if rep.Status != api.StatusOK {
		return wrapAPIError(rep.Response)
	}
	return nil
}

// ProcessingSuccess signal the success of a processing
// for actors with no outputs of it not all outputs were
// hit or closed
func (a *Actor) ProcessingSuccess(envelope *api.Envelope) error {
	return a.ProcessingEnd(envelope, api.ActorProcessingSuccess, nil)
}

// ProcessingError signals a processing error.
func (a *Actor) ProcessingError(envelope *api.Envelope, format string, fmtArgs ...interface{}) error {
	return a.ProcessingEnd(envelope, api.ActorProcessingError, []api.LogMessage{
		api.LogMessage{
			Time:  time.Now(),
			Level: api.LogError,
			Text:  fmt.Sprintf(format, fmtArgs...),
		}})
}

// SendEnvelope send a potentially incomplete envelope to the broker
func (a *Actor) SendEnvelope(output string, env api.Envelope) (status api.ReceptionStatus, err error) {
	if a.Kind == api.ConsumerKind {
		return api.ReceptionError, fmt.Errorf("A 'Consumer' actor cannot emit to outputs")
	}
	return a.Client.SendEnvelope(a.ID, output, env)
}

// SendItems send a group of items to the broker
// The envelope and event will be constructed automatically
func (a *Actor) SendItems(output, eventType string, items []api.Item) (status api.ReceptionStatus, err error) {
	return a.SendEnvelope(output, ForgeEnvelopeFromItems(eventType, items))
}

// SendItemsContext send a group of items to the broker, in a given context
// The envelope and event will be constructed automatically
func (a *Actor) SendItemsContext(context api.EnvelopeContext, output, eventType string, items []api.Item) (status api.ReceptionStatus, err error) {
	envelope := ForgeEnvelopeFromItems(eventType, items)
	envelope.Context = context
	return a.SendEnvelope(output, envelope)
}

// ForgeEnvelopeFromEvent ...
func ForgeEnvelopeFromEvent(event api.Event) api.Envelope {
	envID, err := uuid.V4()
	if err != nil {
		panic(err)
	}
	return api.Envelope{
		ID:       envID,
		EventIDs: []uuid.UUID{event.ID},
		Events:   []api.Event{event},
	}
}

// ForgeEnvelopeFromItems ...
func ForgeEnvelopeFromItems(eventType string, items []api.Item) api.Envelope {
	evID, err := uuid.V4()
	if err != nil {
		panic(err)
	}
	checksum := api.Checksum(0).
		Update(evID[:16]).
		Update([]byte(eventType))

	for _, item := range items {
		checksum = checksum.Update(item)
	}

	event := api.Event{
		ID:        evID,
		Type:      eventType,
		Index:     1,
		Items:     items,
		ItemCount: len(items),
		Checksum:  checksum,
	}
	return ForgeEnvelopeFromEvent(event)
}

// Subscribe setup the subscription(s)
func (a *Actor) Subscribe(handler EnvelopeHandler) (*Subscription, error) {
	if a.Kind == api.EmitterKind {
		return nil, fmt.Errorf("A 'Emitter' actor cannot subscribe to inputs")
	}
	subject := api.ActorInputSubject(a.ID, api.DefaultInput)
	a.GetClient().Log.Noticef("Subscribing to: %s", subject)
	s := Subscription{
		actor:   a,
		handler: handler,
	}
	sub, err := a.Client.conn.QueueSubscribe(
		subject,
		a.ID.String(),
		s.OnReceiveEnvelope)
	if err != nil {
		return nil, err
	}
	s.sub = sub
	a.GetClient().Log.Noticef("Listening to: %s", subject)
	if err := a.Client.conn.Publish(
		api.ActorReadySubject(a.ID), api.ActorReadyMsg{},
	); err != nil {

	}
	return &s, nil
}

// Subscription holds the nats subcription and call the handler when
// an envelope starts arriving
type Subscription struct {
	actor   *Actor
	handler EnvelopeHandler
	sub     *nats.Subscription
}

// Unsubscribe unsuscribes the nats handler
func (s *Subscription) Unsubscribe() error {
	return s.sub.Unsubscribe()
}

// OnReceiveEnvelope is the handler for incoming envelopes
func (s *Subscription) OnReceiveEnvelope(subject, reply string, e api.Envelope) {
	s.actor.GetClient().Log.Noticef("Received Envelope")
	sSubject := strings.Split(subject, ".")
	input := sSubject[len(sSubject)-1]
	v, err := tools.NewEnvelopeValidator(e)
	if err != nil {
		s.actor.Log.Errorf("Validator error: %s", err)
		return
	}
	if !v.Complete {
		err = fmt.Errorf("Not implemented: incomplete envelopes")
	} else if !v.Valid {
		err = fmt.Errorf("Invalid envelope")
	}

	var ack api.EnvelopeAck
	if err != nil {
		ack = api.NewEnvelopeAckError(e.ID, err)
	} else {
		s.handler(input, NewEnvelopeReader(e))
		ack = api.EnvelopeAck{
			ID:              e.ID,
			ReceptionStatus: api.ReceptionAccepted,
		}
	}
	// XXX this reply should be handled by the reader
	if err := s.actor.Client.conn.Publish(reply, ack); err != nil {
		s.actor.GetClient().Log.Errorf("Error publishing reply: %s", err)
		return
	}
	s.actor.GetClient().Log.Noticef("Replied on %s", reply)
}
